"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZnifferLogger = exports.ZNIFFER_LABEL = void 0;
const cc_1 = require("@zwave-js/cc");
const core_1 = require("@zwave-js/core");
const shared_1 = require("@zwave-js/shared");
const strings_1 = require("alcalzone-shared/strings");
exports.ZNIFFER_LABEL = "ZNIFFR";
const ZNIFFER_LOGLEVEL = "info";
class ZnifferLogger extends core_1.ZWaveLoggerBase {
    zniffer;
    constructor(zniffer, loggers) {
        super(loggers, exports.ZNIFFER_LABEL);
        this.zniffer = zniffer;
    }
    isLogVisible() {
        return this.container.isLoglevelVisible(ZNIFFER_LOGLEVEL);
    }
    /**
     * Logs a message
     * @param msg The message to output
     */
    print(message, level) {
        const actualLevel = level || ZNIFFER_LOGLEVEL;
        if (!this.container.isLoglevelVisible(actualLevel))
            return;
        this.logger.log({
            level: actualLevel,
            message,
            direction: (0, core_1.getDirectionPrefix)("none"),
            context: { source: "zniffer", direction: "none" },
        });
    }
    crcError(frame, rssi) {
        if (!this.isLogVisible())
            return;
        const logEntry = {
            tags: ["CRC ERROR"],
            message: {
                channel: frame.channel,
                "protocol/data rate": (0, core_1.znifferProtocolDataRateToString)(frame.protocolDataRate),
                RSSI: rssi != undefined
                    ? (0, core_1.rssiToString)(rssi)
                    : frame.rssiRaw.toString(),
                payload: (0, shared_1.buffer2hex)(frame.payload),
            },
        };
        const msg = [(0, core_1.tagify)(logEntry.tags)];
        msg.push(...(0, core_1.messageRecordToLines)(logEntry.message).map((line) => "  " + line));
        try {
            // If possible, include information about the CCs
            this.logger.log({
                level: "warn",
                message: msg,
                direction: (0, core_1.getDirectionPrefix)("inbound"),
                context: { source: "zniffer", direction: "inbound" },
            });
        }
        catch { }
    }
    mpdu(mpdu, payloadCC) {
        if (!this.isLogVisible())
            return;
        const hasPayload = !!payloadCC || mpdu.payload.length > 0;
        const logEntry = mpdu.toLogEntry();
        let msg = [(0, core_1.tagify)(logEntry.tags)];
        if (logEntry.message) {
            msg.push(...(0, core_1.messageRecordToLines)(logEntry.message).map((line) => (hasPayload ? "│ " : "  ") + line));
        }
        try {
            // If possible, include information about the CCs
            if (!!payloadCC) {
                // Remove the default payload message and draw a bracket
                msg = msg.filter((line) => !line.startsWith("│ payload:"));
                const logCC = (cc, indent = 0) => {
                    const isEncapCC = (0, cc_1.isEncapsulatingCommandClass)(cc)
                        || (0, cc_1.isMultiEncapsulatingCommandClass)(cc);
                    const loggedCC = cc.toLogEntry(this.zniffer);
                    msg.push(" ".repeat(indent * 2) + "└─" + (0, core_1.tagify)(loggedCC.tags));
                    indent++;
                    if (loggedCC.message) {
                        msg.push(...(0, core_1.messageRecordToLines)(loggedCC.message).map((line) => `${" ".repeat(indent * 2)}${isEncapCC ? "│ " : "  "}${line}`));
                    }
                    // If this is an encap CC, continue
                    if ((0, cc_1.isEncapsulatingCommandClass)(cc)) {
                        logCC(cc.encapsulated, indent);
                    }
                    else if ((0, cc_1.isMultiEncapsulatingCommandClass)(cc)) {
                        for (const encap of cc.encapsulated) {
                            logCC(encap, indent);
                        }
                    }
                };
                logCC(payloadCC);
            }
            const homeId = (0, strings_1.padStart)(mpdu.homeId.toString(16), 8, "0")
                .toLowerCase();
            this.logger.log({
                level: ZNIFFER_LOGLEVEL,
                secondaryTags: (0, core_1.tagify)([homeId]),
                message: msg,
                direction: (0, core_1.getDirectionPrefix)("inbound"),
                context: { source: "zniffer", direction: "inbound" },
            });
        }
        catch { }
    }
    beam(beam) {
        if (!this.isLogVisible())
            return;
        const logEntry = beam.toLogEntry();
        const msg = [(0, core_1.tagify)(logEntry.tags)];
        if (logEntry.message) {
            msg.push(...(0, core_1.messageRecordToLines)(logEntry.message).map((line) => ("  ") + line));
        }
        try {
            // If possible, include information about the CCs
            let secondaryTags;
            if ("homeIdHash" in beam && beam.homeIdHash) {
                secondaryTags = (0, core_1.tagify)([`hash: ${(0, shared_1.num2hex)(beam.homeIdHash)}`]);
            }
            this.logger.log({
                level: ZNIFFER_LOGLEVEL,
                secondaryTags,
                message: msg,
                direction: (0, core_1.getDirectionPrefix)("inbound"),
                context: { source: "zniffer", direction: "inbound" },
            });
        }
        catch { }
    }
}
exports.ZnifferLogger = ZnifferLogger;
//# sourceMappingURL=Zniffer.js.map