"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.znifferDataMessageToCorruptedFrame = exports.beamToFrame = exports.mpduToLongRangeFrame = exports.mpduToZWaveFrame = exports.mpduToFrame = exports.BeamStop = exports.LongRangeBeamStart = exports.ZWaveBeamStart = exports.parseBeamFrame = exports.SearchResultExplorerZWaveMPDU = exports.InclusionRequestExplorerZWaveMPDU = exports.NormalExplorerZWaveMPDU = exports.ExplorerZWaveMPDU = exports.MulticastZWaveMPDU = exports.RoutedZWaveMPDU = exports.AckZWaveMPDU = exports.SinglecastZWaveMPDU = exports.ZWaveMPDU = exports.AckLongRangeMPDU = exports.SinglecastLongRangeMPDU = exports.LongRangeMPDU = exports.parseMPDU = void 0;
const core_1 = require("@zwave-js/core");
const serial_1 = require("@zwave-js/serial");
const shared_1 = require("@zwave-js/shared");
const strings_1 = require("alcalzone-shared/strings");
const SendDataShared_1 = require("../serialapi/transport/SendDataShared");
const _Types_1 = require("./_Types");
function getChannelConfiguration(region) {
    switch (region) {
        case core_1.ZnifferRegion.Japan:
        case core_1.ZnifferRegion.Korea:
            return "3";
        case core_1.ZnifferRegion["USA (Long Range)"]:
        case core_1.ZnifferRegion["USA (Long Range, backup)"]:
        case core_1.ZnifferRegion["USA (Long Range, end device)"]:
            return "4";
        default:
            return "1/2";
    }
}
function longRangeBeamPowerToDBm(power) {
    return [
        -6,
        -2,
        2,
        6,
        10,
        13,
        16,
        19,
        21,
        23,
        25,
        26,
        27,
        28,
        29,
        30,
    ][power];
}
function formatNodeId(nodeId) {
    return (0, strings_1.padStart)(nodeId.toString(), 3, "0");
}
function formatRoute(source, repeaters, destination, direction, currentHop, failedHop) {
    return [
        direction === "outbound"
            ? formatNodeId(source)
            : formatNodeId(destination),
        ...repeaters.map(formatNodeId),
        direction === "outbound"
            ? formatNodeId(destination)
            : formatNodeId(source),
    ].map((id, i) => {
        if (i === 0)
            return id;
        if (i - 1 === failedHop)
            return " × " + id;
        if (i - 1 === currentHop) {
            return (direction === "outbound" ? " » " : " « ") + id;
        }
        return (direction === "outbound" ? " › " : " ‹ ") + id;
    })
        .join("");
}
function parseMPDU(frame) {
    switch (frame.channel) {
        case 0:
        case 1:
        case 2:
            return ZWaveMPDU.from(frame);
        case 3:
            return LongRangeMPDU.from(frame);
        default:
            core_1.validatePayload.fail(`Unsupported channel ${frame.channel}. MPDU payload: ${(0, shared_1.buffer2hex)(frame.payload)}`);
    }
}
exports.parseMPDU = parseMPDU;
class LongRangeMPDU {
    constructor(options) {
        const data = options.data;
        this.frameInfo = options.frameInfo;
        if (options.frameInfo.channel !== 3) {
            core_1.validatePayload.fail(`Unsupported channel ${options.frameInfo.channel} for LongRangeMPDU`);
        }
        this.homeId = data.readUInt32BE(0);
        const nodeIds = data.readUIntBE(4, 3);
        this.sourceNodeId = nodeIds >>> 12;
        this.destinationNodeId = nodeIds & 0xfff;
        // skip length byte
        const frameControl = data[8];
        this.ackRequested = !!(frameControl & 0b1000_0000);
        const hasExtendedHeader = !!(frameControl & 0b0100_0000);
        this.headerType = frameControl & 0b0000_0111;
        this.sequenceNumber = data[9];
        this.noiseFloor = (0, SendDataShared_1.parseRSSI)(data, 10);
        this.txPower = data.readInt8(11);
        let offset = 12;
        if (hasExtendedHeader) {
            const extensionControl = data[offset++];
            const extensionLength = extensionControl & 0b111;
            // const discardUnknown = extensionControl & 0b0000_1000;
            // const extensionType = (extensionControl & 0b0111_0000) >>> 4;
            // TODO: Parse extension (once there is a definition)
            offset += extensionLength;
        }
        const Constructor = this.headerType === core_1.MPDUHeaderType.Acknowledgement
            ? AckLongRangeMPDU
            : this.headerType === core_1.MPDUHeaderType.Singlecast
                ? SinglecastLongRangeMPDU
                : undefined;
        if (!Constructor) {
            core_1.validatePayload.fail(`Unsupported Long Range MPDU header type ${this.headerType}`);
        }
        else if (new.target !== Constructor
            && !(0, shared_1.staticExtends)(new.target, Constructor)) {
            return new Constructor(options);
        }
        this.payload = data.subarray(offset);
    }
    frameInfo;
    homeId;
    sourceNodeId;
    destinationNodeId;
    ackRequested;
    headerType;
    sequenceNumber;
    noiseFloor;
    txPower;
    payload;
    static from(msg) {
        return new LongRangeMPDU({
            data: msg.payload,
            frameInfo: (0, shared_1.pick)(msg, [
                "channel",
                "frameType",
                "region",
                "protocolDataRate",
                "rssiRaw",
            ]),
        });
    }
    toLogEntry() {
        const tags = [
            formatRoute(this.sourceNodeId, [], this.destinationNodeId, 
            // Singlecast frames do not contain a bit for this, we consider them all "outbound"
            "outbound", 0),
        ];
        if (this.headerType === core_1.MPDUHeaderType.Acknowledgement) {
            tags.unshift("ACK");
        }
        const message = {
            "sequence no.": this.sequenceNumber,
            channel: this.frameInfo.channel,
            "protocol/data rate": (0, core_1.znifferProtocolDataRateToString)(this.frameInfo.protocolDataRate),
            "TX power": `${this.txPower} dBm`,
            RSSI: this.frameInfo.rssi != undefined
                ? (0, core_1.rssiToString)(this.frameInfo.rssi)
                : this.frameInfo.rssiRaw.toString(),
            "noise floor": (0, core_1.rssiToString)(this.noiseFloor),
        };
        if (this.headerType !== core_1.MPDUHeaderType.Acknowledgement) {
            message["ack requested"] = this.ackRequested;
        }
        if (this.payload.length > 0) {
            message.payload = (0, shared_1.buffer2hex)(this.payload);
        }
        return {
            tags,
            message,
        };
    }
}
exports.LongRangeMPDU = LongRangeMPDU;
class SinglecastLongRangeMPDU extends LongRangeMPDU {
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        const message = {
            ...original,
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.SinglecastLongRangeMPDU = SinglecastLongRangeMPDU;
class AckLongRangeMPDU extends LongRangeMPDU {
    constructor(options) {
        super(options);
        this.incomingRSSI = (0, SendDataShared_1.parseRSSI)(this.payload, 0);
        this.payload = this.payload.subarray(1);
    }
    incomingRSSI;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        const message = {
            ...original,
            "incoming RSSI": (0, core_1.rssiToString)(this.incomingRSSI),
        };
        if (this.payload.length > 0) {
            message.payload = (0, shared_1.buffer2hex)(this.payload);
        }
        return {
            tags,
            message,
        };
    }
}
exports.AckLongRangeMPDU = AckLongRangeMPDU;
class ZWaveMPDU {
    constructor(options) {
        const data = options.data;
        this.frameInfo = options.frameInfo;
        let destinationOffset = 8;
        const frameControl = data.subarray(5, 7);
        switch (options.frameInfo.channel) {
            case 0:
            case 1: {
                this.routed = !!(frameControl[0] & 0b1000_0000);
                this.ackRequested = !!(frameControl[0] & 0b0100_0000);
                this.lowPower = !!(frameControl[0] & 0b0010_0000);
                this.speedModified = !!(frameControl[0] & 0b0001_0000);
                this.headerType = frameControl[0] & 0b0000_1111;
                this.beamingInfo = frameControl[1] & 0b0110_0000;
                this.sequenceNumber = frameControl[1] & 0b0000_1111;
                break;
            }
            case 2: {
                this.routed = false;
                this.ackRequested = !!(frameControl[0] & 0b1000_0000);
                this.lowPower = !!(frameControl[0] & 0b0100_0000);
                this.speedModified = false;
                this.headerType = frameControl[0] & 0b0000_1111;
                this.beamingInfo = frameControl[1] & 0b0111_0000;
                this.sequenceNumber = data[destinationOffset];
                destinationOffset++;
                break;
            }
            case 3: {
                core_1.validatePayload.fail(`Channel 3 (ZWLR) must be parsed as a LongRangeMPDU!`);
            }
            default: {
                core_1.validatePayload.fail(`Unsupported channel ${options.frameInfo.channel}. MPDU payload: ${(0, shared_1.buffer2hex)(data)}`);
            }
        }
        const Constructor = this.headerType === core_1.MPDUHeaderType.Acknowledgement
            ? AckZWaveMPDU
            : (this.headerType === core_1.MPDUHeaderType.Routed
                || (this.headerType === core_1.MPDUHeaderType.Singlecast
                    && this.routed))
                ? RoutedZWaveMPDU
                : this.headerType === core_1.MPDUHeaderType.Singlecast
                    ? SinglecastZWaveMPDU
                    : this.headerType === core_1.MPDUHeaderType.Multicast
                        ? MulticastZWaveMPDU
                        : this.headerType === core_1.MPDUHeaderType.Explorer
                            ? ExplorerZWaveMPDU
                            : undefined;
        if (!Constructor) {
            core_1.validatePayload.fail(`Unsupported MPDU header type ${this.headerType}`);
        }
        else if (new.target !== Constructor
            && !(0, shared_1.staticExtends)(new.target, Constructor)) {
            return new Constructor(options);
        }
        // FIXME: Parse Beams
        this.homeId = data.readUInt32BE(0);
        this.sourceNodeId = data[4];
        // byte 7 is another length byte
        // FIXME: This should consider the multicast control byte
        const destinationLength = this.headerType === core_1.MPDUHeaderType.Multicast
            ? 30
            : 1;
        this.destinationBuffer = data.subarray(destinationOffset, destinationOffset + destinationLength);
        this.payload = data.subarray(destinationOffset + destinationLength);
    }
    frameInfo;
    homeId;
    sourceNodeId;
    routed;
    ackRequested;
    lowPower;
    speedModified;
    headerType;
    beamingInfo;
    sequenceNumber;
    destinationBuffer;
    payload;
    static from(msg) {
        return new ZWaveMPDU({
            data: msg.payload,
            frameInfo: (0, shared_1.pick)(msg, [
                "channel",
                "frameType",
                "region",
                "protocolDataRate",
                "rssiRaw",
            ]),
        });
    }
    toLogEntry() {
        const tags = [formatNodeId(this.sourceNodeId)];
        const message = {
            "sequence no.": this.sequenceNumber,
            channel: this.frameInfo.channel,
            "protocol/data rate": (0, core_1.znifferProtocolDataRateToString)(this.frameInfo.protocolDataRate)
                + (this.speedModified ? " (reduced)" : ""),
            RSSI: this.frameInfo.rssi != undefined
                ? (0, core_1.rssiToString)(this.frameInfo.rssi)
                : this.frameInfo.rssiRaw.toString(),
        };
        return {
            tags,
            message,
        };
    }
}
exports.ZWaveMPDU = ZWaveMPDU;
class SinglecastZWaveMPDU extends ZWaveMPDU {
    constructor(options) {
        super(options);
        this.destinationNodeId = this.destinationBuffer[0];
    }
    destinationNodeId;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, [], this.destinationNodeId, 
        // Singlecast frames do not contain a bit for this, we consider them all "outbound"
        "outbound", 0);
        const message = {
            ...original,
            "ack requested": this.ackRequested,
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.SinglecastZWaveMPDU = SinglecastZWaveMPDU;
class AckZWaveMPDU extends ZWaveMPDU {
    constructor(options) {
        super(options);
        this.destinationNodeId = this.destinationBuffer[0];
    }
    destinationNodeId;
    toLogEntry() {
        const { tags, message } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, [], this.destinationNodeId, 
        // ACK frames do not contain a bit for this, we consider them all "inbound"
        "inbound", 0);
        tags.unshift("ACK");
        return {
            tags,
            message,
        };
    }
}
exports.AckZWaveMPDU = AckZWaveMPDU;
class RoutedZWaveMPDU extends ZWaveMPDU {
    constructor(options) {
        super(options);
        const channelConfig = getChannelConfiguration(this.frameInfo.region);
        this.direction = (this.payload[0] & 0b1) ? "inbound" : "outbound";
        this.routedAck = !!(this.payload[0] & 0b10);
        this.routedError = !!(this.payload[0] & 0b100);
        const hasExtendedHeader = !!(this.payload[0] & 0b1000);
        if (this.routedError) {
            this.failedHop = this.payload[0] >>> 4;
        }
        else if (channelConfig === "1/2") {
            // @ts-expect-error speedModified is readonly
            this.speedModified = !!(this.payload[0] & 0b10000);
        }
        this.hop = this.payload[1] & 0b1111;
        // The hop field in the MPDU indicates which repeater should handle the frame next.
        // This means that for an inbound frame between repeater 0 and 1, the value is one
        // less (0) than for an outbound frame (1). This also means that the field overflows
        // to 0x0f when the frame returns to the source node.
        //
        // We normalize this, so hop = 0 always means the frame is transmitted between the source node and repeater 0.
        if (this.direction === "inbound") {
            this.hop = (this.hop + 1) % 16;
        }
        const numRepeaters = this.payload[1] >>> 4;
        this.repeaters = [...this.payload.subarray(2, 2 + numRepeaters)];
        let offset = 2 + numRepeaters;
        if (channelConfig === "3") {
            this.destinationWakeup = this.payload[offset++] === 0x02;
        }
        if (hasExtendedHeader) {
            const headerPreamble = this.payload[offset++];
            const headerLength = headerPreamble >>> 4;
            const headerType = headerPreamble & 0b1111;
            const header = this.payload.subarray(offset, offset + headerLength);
            offset += headerLength;
            if (headerType === 0x00) {
                this.destinationWakeupType = header[0] & 0b0100_0000
                    ? "1000ms"
                    : header[0] & 0b0010_0000
                        ? "250ms"
                        : undefined;
            }
            else if (headerType === 0x01) {
                const repeaterRSSI = [];
                for (let i = 0; i < numRepeaters; i++) {
                    repeaterRSSI.push((0, SendDataShared_1.parseRSSI)(header, i));
                }
                this.repeaterRSSI = repeaterRSSI;
            }
        }
        this.payload = this.payload.subarray(offset);
        this.destinationNodeId = this.destinationBuffer[0];
    }
    destinationNodeId;
    direction;
    routedAck;
    routedError;
    failedHop;
    hop;
    repeaters;
    destinationWakeup;
    destinationWakeupType;
    repeaterRSSI;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, this.repeaters, this.destinationNodeId, this.direction, this.hop, this.failedHop);
        const message = {
            ...original,
            "ack requested": this.ackRequested,
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.RoutedZWaveMPDU = RoutedZWaveMPDU;
class MulticastZWaveMPDU extends ZWaveMPDU {
    constructor(options) {
        super(options);
        const control = this.destinationBuffer[0]; // 3 bits offset, 5 bits mask length, but this MUST be set to 29
        core_1.validatePayload.withReason("Invalid multicast control byte")(control === 29);
        this.destinationNodeIds = (0, core_1.parseNodeBitMask)(this.destinationBuffer.subarray(1));
    }
    destinationNodeIds;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags.push("MULTICAST");
        const message = {
            destinations: this.destinationNodeIds.join(", "),
            ...original,
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.MulticastZWaveMPDU = MulticastZWaveMPDU;
class ExplorerZWaveMPDU extends ZWaveMPDU {
    constructor(options) {
        super(options);
        this.version = this.payload[0] >>> 5;
        this.command = this.payload[0] & 0b0001_1111;
        this.stop = !!(this.payload[1] & 0b100);
        this.direction = this.payload[1] & 0b010 ? "inbound" : "outbound";
        this.sourceRouted = !!(this.payload[1] & 0b001);
        this.randomTXInterval = this.payload[2];
        this.ttl = this.payload[3] >>> 4;
        const numRepeaters = this.payload[3] & 0b1111;
        this.repeaters = [...this.payload.subarray(4, 4 + numRepeaters)];
        // Make sure the correct constructor gets called
        const Constructor = this.command === _Types_1.ExplorerFrameCommand.Normal
            ? NormalExplorerZWaveMPDU
            : this.command === _Types_1.ExplorerFrameCommand.InclusionRequest
                ? InclusionRequestExplorerZWaveMPDU
                : this.command === _Types_1.ExplorerFrameCommand.SearchResult
                    ? SearchResultExplorerZWaveMPDU
                    : undefined;
        if (!Constructor) {
            core_1.validatePayload.fail(`Unsupported Explorer MPDU command ${this.command}`);
        }
        else if (new.target !== Constructor
            && !(0, shared_1.staticExtends)(new.target, Constructor)) {
            return new Constructor(options);
        }
        this.destinationNodeId = this.destinationBuffer[0];
        this.payload = this.payload.subarray(8);
    }
    destinationNodeId;
    version;
    command;
    stop;
    sourceRouted;
    direction;
    randomTXInterval;
    ttl;
    repeaters;
}
exports.ExplorerZWaveMPDU = ExplorerZWaveMPDU;
class NormalExplorerZWaveMPDU extends ExplorerZWaveMPDU {
    constructor(options) {
        super(options);
    }
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, this.repeaters, this.destinationNodeId, 
        // Explorer frames do not contain a bit for the direction, we consider them all "outbound"
        "outbound", 4 - this.ttl);
        tags.unshift("EXPLORER");
        const message = {
            ...original,
            "ack requested": this.ackRequested,
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.NormalExplorerZWaveMPDU = NormalExplorerZWaveMPDU;
class InclusionRequestExplorerZWaveMPDU extends ExplorerZWaveMPDU {
    constructor(options) {
        super(options);
        this.networkHomeId = this.payload.readUInt32BE(0);
        this.payload = this.payload.subarray(4);
    }
    /** The home ID of the repeating node */
    networkHomeId;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, this.repeaters, this.destinationNodeId, 
        // Explorer frames do not contain a bit for the direction, we consider them all "outbound"
        "outbound", 4 - this.ttl);
        tags.unshift("INCL REQUEST");
        const message = {
            ...original,
            "network home ID": (0, strings_1.padStart)(this.networkHomeId.toString(16), 8, "0"),
            payload: (0, shared_1.buffer2hex)(this.payload),
        };
        return {
            tags,
            message,
        };
    }
}
exports.InclusionRequestExplorerZWaveMPDU = InclusionRequestExplorerZWaveMPDU;
class SearchResultExplorerZWaveMPDU extends ExplorerZWaveMPDU {
    constructor(options) {
        super(options);
        this.searchingNodeId = this.payload[0];
        this.frameHandle = this.payload[1];
        this.resultTTL = this.payload[2] >>> 4;
        const numRepeaters = this.payload[2] & 0b1111;
        this.resultRepeaters = [
            ...this.payload.subarray(3, 3 + numRepeaters),
        ];
        // This frame contains no payload
        this.payload = Buffer.allocUnsafe(0);
    }
    /** The node ID that sent the explorer frame that's being answered here */
    searchingNodeId;
    /** The sequence number of the original explorer frame */
    frameHandle;
    resultTTL;
    resultRepeaters;
    toLogEntry() {
        const { tags, message: original } = super.toLogEntry();
        tags[0] = formatRoute(this.sourceNodeId, this.repeaters, this.destinationNodeId, 
        // Explorer frames do not contain a bit for the direction, we consider their responses "inbound"
        "inbound", 4 - this.ttl);
        tags.unshift("EXPLORER RESULT");
        const message = {
            ...original,
            "frame handle": this.frameHandle,
            "result TTL": this.resultTTL,
            "result repeaters": this.resultRepeaters.join(", "),
        };
        return {
            tags,
            message,
        };
    }
}
exports.SearchResultExplorerZWaveMPDU = SearchResultExplorerZWaveMPDU;
function parseBeamFrame(frame) {
    if (frame.frameType === serial_1.ZnifferFrameType.BeamStop) {
        return new BeamStop({
            data: frame.payload,
            frameInfo: frame,
        });
    }
    const channelConfig = getChannelConfiguration(frame.region);
    switch (channelConfig) {
        case "1/2":
        case "3": {
            return new ZWaveBeamStart({
                data: frame.payload,
                frameInfo: frame,
            });
        }
        case "4": {
            return new LongRangeBeamStart({
                data: frame.payload,
                frameInfo: frame,
            });
        }
        default:
            core_1.validatePayload.fail(
            // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
            `Unsupported channel configuration ${channelConfig}. MPDU payload: ${(0, shared_1.buffer2hex)(frame.payload)}`);
    }
}
exports.parseBeamFrame = parseBeamFrame;
class ZWaveBeamStart {
    constructor(options) {
        const data = options.data;
        this.frameInfo = options.frameInfo;
        const channelConfig = getChannelConfiguration(this.frameInfo.region);
        switch (channelConfig) {
            case "1/2":
            case "3":
                // OK
                break;
            case "4": {
                core_1.validatePayload.fail(`Channel configuration 4 (ZWLR) must be parsed as a LongRangeMPDU!`);
            }
            default: {
                core_1.validatePayload.fail(
                // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
                `Unsupported channel configuration ${channelConfig}. MPDU payload: ${(0, shared_1.buffer2hex)(data)}`);
            }
        }
        this.destinationNodeId = data[1];
        if (data[2] === 0x01) {
            this.homeIdHash = data[3];
        }
    }
    frameInfo;
    homeIdHash;
    destinationNodeId;
    toLogEntry() {
        const tags = [
            `BEAM » ${formatNodeId(this.destinationNodeId)}`,
        ];
        const message = {
            channel: this.frameInfo.channel,
            "protocol/data rate": (0, core_1.znifferProtocolDataRateToString)(this.frameInfo.protocolDataRate),
            RSSI: this.frameInfo.rssi != undefined
                ? (0, core_1.rssiToString)(this.frameInfo.rssi)
                : this.frameInfo.rssiRaw.toString(),
        };
        return {
            tags,
            message,
        };
    }
}
exports.ZWaveBeamStart = ZWaveBeamStart;
class LongRangeBeamStart {
    constructor(options) {
        const data = options.data;
        this.frameInfo = options.frameInfo;
        const channelConfig = getChannelConfiguration(this.frameInfo.region);
        switch (channelConfig) {
            case "1/2":
            case "3":
                // OK
                break;
            case "4": {
                core_1.validatePayload.fail(`Channel configuration 4 (ZWLR) must be parsed as a LongRangeMPDU!`);
            }
            default: {
                core_1.validatePayload.fail(
                // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
                `Unsupported channel configuration ${channelConfig}. MPDU payload: ${(0, shared_1.buffer2hex)(data)}`);
            }
        }
        const txPower = data[1] >>> 4;
        this.txPower = longRangeBeamPowerToDBm(txPower);
        this.destinationNodeId = data.readUint16BE(1) & 0x0fff;
        this.homeIdHash = data[3];
    }
    frameInfo;
    homeIdHash;
    destinationNodeId;
    txPower;
    toLogEntry() {
        const tags = [
            `BEAM » ${formatNodeId(this.destinationNodeId)}`,
        ];
        const message = {
            channel: this.frameInfo.channel,
            "protocol/data rate": (0, core_1.znifferProtocolDataRateToString)(this.frameInfo.protocolDataRate),
            "TX power": `${this.txPower} dBm`,
            RSSI: this.frameInfo.rssi != undefined
                ? (0, core_1.rssiToString)(this.frameInfo.rssi)
                : this.frameInfo.rssiRaw.toString(),
        };
        return {
            tags,
            message,
        };
    }
}
exports.LongRangeBeamStart = LongRangeBeamStart;
class BeamStop {
    constructor(options) {
        this.frameInfo = options.frameInfo;
    }
    frameInfo;
    toLogEntry() {
        const tags = [
            "BEAM STOP",
        ];
        const message = {
            channel: this.frameInfo.channel,
        };
        return {
            tags,
            message,
        };
    }
}
exports.BeamStop = BeamStop;
function mpduToFrame(mpdu, payloadCC) {
    if (mpdu instanceof ZWaveMPDU) {
        return mpduToZWaveFrame(mpdu, payloadCC);
    }
    else if (mpdu instanceof LongRangeMPDU) {
        return mpduToLongRangeFrame(mpdu, payloadCC);
    }
    throw new core_1.ZWaveError(`mpduToFrame not supported for ${mpdu.constructor.name}`, core_1.ZWaveErrorCodes.Argument_Invalid);
}
exports.mpduToFrame = mpduToFrame;
function mpduToZWaveFrame(mpdu, payloadCC) {
    const retBase = {
        protocol: core_1.Protocols.ZWave,
        channel: mpdu.frameInfo.channel,
        region: mpdu.frameInfo.region,
        rssiRaw: mpdu.frameInfo.rssiRaw,
        rssi: mpdu.frameInfo.rssi,
        protocolDataRate: mpdu.frameInfo.protocolDataRate,
        speedModified: mpdu.speedModified,
        sequenceNumber: mpdu.sequenceNumber,
        homeId: mpdu.homeId,
        sourceNodeId: mpdu.sourceNodeId,
    };
    if (mpdu instanceof SinglecastZWaveMPDU) {
        const ret = {
            ...retBase,
            ackRequested: mpdu.ackRequested,
            payload: payloadCC ?? mpdu.payload,
        };
        if (mpdu.destinationNodeId === core_1.NODE_ID_BROADCAST) {
            return {
                type: _Types_1.ZWaveFrameType.Broadcast,
                destinationNodeId: mpdu.destinationNodeId,
                ...ret,
            };
        }
        else {
            return {
                type: _Types_1.ZWaveFrameType.Singlecast,
                destinationNodeId: mpdu.destinationNodeId,
                ...ret,
            };
        }
    }
    else if (mpdu instanceof AckZWaveMPDU) {
        return {
            type: _Types_1.ZWaveFrameType.AckDirect,
            ...retBase,
            destinationNodeId: mpdu.destinationNodeId,
        };
    }
    else if (mpdu instanceof MulticastZWaveMPDU) {
        return {
            type: _Types_1.ZWaveFrameType.Multicast,
            ...retBase,
            destinationNodeIds: [...mpdu.destinationNodeIds],
            payload: payloadCC ?? mpdu.payload,
        };
    }
    else if (mpdu instanceof RoutedZWaveMPDU) {
        return {
            type: _Types_1.ZWaveFrameType.Singlecast,
            ...retBase,
            destinationNodeId: mpdu.destinationNodeId,
            ackRequested: mpdu.ackRequested,
            payload: payloadCC ?? mpdu.payload,
            direction: mpdu.direction,
            hop: mpdu.hop,
            repeaters: [...mpdu.repeaters],
            repeaterRSSI: mpdu.repeaterRSSI && [...mpdu.repeaterRSSI],
            routedAck: mpdu.routedAck,
            routedError: mpdu.routedError,
            failedHop: mpdu.failedHop,
        };
    }
    else if (mpdu instanceof ExplorerZWaveMPDU) {
        const explorerBase = {
            ...retBase,
            destinationNodeId: mpdu.destinationNodeId,
            ackRequested: mpdu.ackRequested,
            direction: mpdu.direction,
            repeaters: [...mpdu.repeaters],
            ttl: mpdu.ttl,
        };
        if (mpdu instanceof NormalExplorerZWaveMPDU) {
            return {
                type: _Types_1.ZWaveFrameType.ExplorerNormal,
                payload: payloadCC ?? mpdu.payload,
                ...explorerBase,
            };
        }
        else if (mpdu instanceof SearchResultExplorerZWaveMPDU) {
            return {
                type: _Types_1.ZWaveFrameType.ExplorerSearchResult,
                ...explorerBase,
                searchingNodeId: mpdu.searchingNodeId,
                frameHandle: mpdu.frameHandle,
                resultTTL: mpdu.resultTTL,
                resultRepeaters: [...mpdu.resultRepeaters],
            };
        }
        else if (mpdu instanceof InclusionRequestExplorerZWaveMPDU) {
            return {
                type: _Types_1.ZWaveFrameType.ExplorerInclusionRequest,
                payload: payloadCC ?? mpdu.payload,
                ...explorerBase,
                networkHomeId: mpdu.networkHomeId,
            };
        }
    }
    throw new core_1.ZWaveError(`mpduToZWaveFrame not supported for ${mpdu.constructor.name}`, core_1.ZWaveErrorCodes.Argument_Invalid);
}
exports.mpduToZWaveFrame = mpduToZWaveFrame;
function mpduToLongRangeFrame(mpdu, payloadCC) {
    const retBase = {
        protocol: core_1.Protocols.ZWaveLongRange,
        channel: mpdu.frameInfo.channel,
        region: mpdu.frameInfo.region,
        protocolDataRate: mpdu.frameInfo.protocolDataRate,
        rssiRaw: mpdu.frameInfo.rssiRaw,
        rssi: mpdu.frameInfo.rssi,
        noiseFloor: mpdu.noiseFloor,
        txPower: mpdu.txPower,
        sequenceNumber: mpdu.sequenceNumber,
        homeId: mpdu.homeId,
        sourceNodeId: mpdu.sourceNodeId,
        destinationNodeId: mpdu.destinationNodeId,
    };
    if (mpdu instanceof SinglecastLongRangeMPDU) {
        const ret = {
            ...retBase,
            ackRequested: mpdu.ackRequested,
            payload: payloadCC ?? mpdu.payload,
        };
        if (mpdu.destinationNodeId === core_1.NODE_ID_BROADCAST_LR) {
            return {
                type: _Types_1.LongRangeFrameType.Broadcast,
                ...ret,
                destinationNodeId: mpdu.destinationNodeId, // Make TS happy
            };
        }
        else {
            return {
                type: _Types_1.LongRangeFrameType.Singlecast,
                ...ret,
            };
        }
    }
    else if (mpdu instanceof AckLongRangeMPDU) {
        return {
            type: _Types_1.LongRangeFrameType.Ack,
            ...retBase,
            incomingRSSI: mpdu.incomingRSSI,
            payload: mpdu.payload,
        };
    }
    throw new core_1.ZWaveError(`mpduToLongRangeFrame not supported for ${mpdu.constructor.name}`, core_1.ZWaveErrorCodes.Argument_Invalid);
}
exports.mpduToLongRangeFrame = mpduToLongRangeFrame;
function beamToFrame(beam) {
    const retBase = {
        channel: beam.frameInfo.channel,
        region: beam.frameInfo.region,
        rssiRaw: beam.frameInfo.rssiRaw,
        rssi: beam.frameInfo.rssi,
        protocolDataRate: beam.frameInfo.protocolDataRate,
    };
    if (beam instanceof ZWaveBeamStart) {
        return {
            protocol: core_1.Protocols.ZWave,
            type: _Types_1.ZWaveFrameType.BeamStart,
            ...retBase,
            destinationNodeId: beam.destinationNodeId,
            homeIdHash: beam.homeIdHash,
        };
    }
    else if (beam instanceof LongRangeBeamStart) {
        return {
            protocol: core_1.Protocols.ZWaveLongRange,
            type: _Types_1.LongRangeFrameType.BeamStart,
            ...retBase,
            destinationNodeId: beam.destinationNodeId,
            homeIdHash: beam.homeIdHash,
            txPower: beam.txPower,
        };
    }
    else {
        // Beam Stop - contains only the channel, the other fields are garbage
        const isLR = beam.frameInfo.channel === 4;
        if (isLR) {
            return {
                protocol: core_1.Protocols.ZWaveLongRange,
                type: _Types_1.LongRangeFrameType.BeamStop,
                channel: beam.frameInfo.channel,
            };
        }
        else {
            return {
                protocol: core_1.Protocols.ZWave,
                type: _Types_1.ZWaveFrameType.BeamStop,
                channel: beam.frameInfo.channel,
            };
        }
    }
}
exports.beamToFrame = beamToFrame;
function znifferDataMessageToCorruptedFrame(msg, rssi) {
    if (msg.checksumOK) {
        throw new core_1.ZWaveError(`znifferDataMessageToCorruptedFrame expects the checksum to be incorrect`, core_1.ZWaveErrorCodes.Argument_Invalid);
    }
    return {
        channel: msg.channel,
        region: msg.region,
        rssiRaw: msg.rssiRaw,
        rssi,
        protocolDataRate: msg.protocolDataRate,
        payload: msg.payload,
    };
}
exports.znifferDataMessageToCorruptedFrame = znifferDataMessageToCorruptedFrame;
//# sourceMappingURL=MPDU.js.map