"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Zniffer = void 0;
const cc_1 = require("@zwave-js/cc");
const core_1 = require("@zwave-js/core");
const serial_1 = require("@zwave-js/serial");
const shared_1 = require("@zwave-js/shared");
const deferred_promise_1 = require("alcalzone-shared/deferred-promise");
const promises_1 = __importDefault(require("node:fs/promises"));
const Zniffer_1 = require("../log/Zniffer");
const CCParsingContext_1 = require("./CCParsingContext");
const MPDU_1 = require("./MPDU");
const logo = `
███████╗ ███╗   ██╗ ██╗ ██████╗ ██████╗ ███████╗ ██████╗          ██╗ ███████╗
╚══███╔╝ ████╗  ██║ ██║ ██╔═══╝ ██╔═══╝ ██╔════╝ ██╔══██╗         ██║ ██╔════╝
  ███╔╝  ██╔██╗ ██║ ██║ ████╗   ████╗   █████╗   ██████╔╝         ██║ ███████╗
 ███╔╝   ██║╚██╗██║ ██║ ██╔═╝   ██╔═╝   ██╔══╝   ██╔══██╗    ██   ██║ ╚════██║
███████╗ ██║ ╚████║ ██║ ██║     ██║     ███████╗ ██║  ██║    ╚█████╔╝ ███████║
╚══════╝ ╚═╝  ╚═══╝ ╚═╝ ╚═╝     ╚═╝     ╚══════╝ ╚═╝  ╚═╝     ╚════╝  ╚══════╝
`.trim();
function is700PlusSeries(chipType) {
    if (typeof chipType !== "string") {
        return chipType.type >= 0x07;
    }
    const chipTypeNumeric = (0, core_1.getChipTypeAndVersion)(chipType);
    if (chipTypeNumeric) {
        return chipTypeNumeric.type >= 0x07;
    }
    return false;
}
function tryConvertRSSI(rssi, chipType) {
    // For 400/500 series, the conversion is documented in the Zniffer user guide.
    // The conversion for 700/800 series was reverse-engineered from the Zniffer firmware.
    // Here, we assume that only these two representations exist:
    if (is700PlusSeries(chipType)) {
        return rssi * 4 - 256;
    }
    else {
        return rssi * 1.5 - 153.5;
    }
}
class Zniffer extends shared_1.TypedEventEmitter {
    port;
    constructor(port, options = {}) {
        super();
        this.port = port;
        // Ensure the given serial port is valid
        if (typeof port !== "string"
            && !(0, serial_1.isZWaveSerialPortImplementation)(port)) {
            throw new core_1.ZWaveError(`The port must be a string or a valid custom serial port implementation!`, core_1.ZWaveErrorCodes.Driver_InvalidOptions);
        }
        // Initialize logging
        this._logContainer = new core_1.ZWaveLogContainer(options.logConfig);
        this.znifferLog = new Zniffer_1.ZnifferLogger(this, this._logContainer);
        this._options = options;
        this._active = false;
    }
    _options;
    /** The serial port instance */
    serial;
    _destroyPromise;
    get wasDestroyed() {
        return !!this._destroyPromise;
    }
    _chipType;
    _currentFrequency;
    /** The currently configured frequency */
    get currentFrequency() {
        return this._currentFrequency;
    }
    _supportedFrequencies = new Map();
    /** A map of supported frequency identifiers and their names */
    get supportedFrequencies() {
        return this._supportedFrequencies;
    }
    _logContainer;
    znifferLog;
    /** The security managers for each node */
    securityManagers = new Map();
    /** A list of awaited messages */
    awaitedMessages = [];
    _active;
    /** Whether the Zniffer instance is currently capturing */
    get active() {
        return this._active;
    }
    _capturedFrames = [];
    /** A list of raw captured frames that can be saved to a .zlf file later */
    get capturedFrames() {
        return this._capturedFrames.filter((f) => f.parsedFrame !== undefined)
            .map((f) => ({
            timestamp: f.timestamp,
            frameData: f.frameData,
            parsedFrame: f.parsedFrame,
        }));
    }
    async init() {
        if (this.wasDestroyed) {
            throw new core_1.ZWaveError("The Zniffer was destroyed. Create a new instance and initialize that one.", core_1.ZWaveErrorCodes.Driver_Destroyed);
        }
        // Open the serial port
        if (typeof this.port === "string") {
            if (this.port.startsWith("tcp://")) {
                const url = new URL(this.port);
                // this.znifferLog.print(`opening serial port ${this.port}`);
                this.serial = new serial_1.ZnifferSocket({
                    host: url.hostname,
                    port: parseInt(url.port),
                }, this._logContainer);
            }
            else {
                // this.znifferLog.print(`opening serial port ${this.port}`);
                this.serial = new serial_1.ZnifferSerialPort(this.port, this._logContainer);
            }
        }
        else {
            // this.znifferLog.print(
            // 	"opening serial port using the provided custom implementation",
            // );
            this.serial = new serial_1.ZnifferSerialPortBase(this.port, this._logContainer);
        }
        this.serial
            .on("data", this.serialport_onData.bind(this))
            .on("error", (err) => {
            this.emit("error", err);
        });
        await this.serial.open();
        this.znifferLog.print(logo, "info");
        await this.stop();
        const versionInfo = await this.getVersion();
        this._chipType = versionInfo.chipType;
        this.znifferLog.print(`received Zniffer info:
  Chip type:       ${typeof versionInfo.chipType === "string"
            ? versionInfo.chipType
            : `unknown (${(0, shared_1.num2hex)(versionInfo.chipType.type)}, ${(0, shared_1.num2hex)(versionInfo.chipType.version)})`}
  Zniffer version: ${versionInfo.majorVersion}.${versionInfo.minorVersion}`, "info");
        await this.setBaudrate(0);
        const freqs = await this.getFrequencies();
        this._currentFrequency = freqs.currentFrequency;
        if (is700PlusSeries(this._chipType)) {
            // The frequencies match the ZnifferRegion enum
            this.znifferLog.print(`received frequency info:
  current frequency:     ${(0, shared_1.getEnumMemberName)(core_1.ZnifferRegion, freqs.currentFrequency)}
  supported frequencies: ${freqs.supportedFrequencies.map((f) => `\n  · ${f.toString().padStart(2, " ")}: ${(0, shared_1.getEnumMemberName)(core_1.ZnifferRegion, f)}`).join("")}`, "info");
            for (const freq of freqs.supportedFrequencies) {
                this._supportedFrequencies.set(freq, (0, shared_1.getEnumMemberName)(core_1.ZnifferRegion, freq));
            }
        }
        else {
            // The frequencies are firmware-specific. Query them from the Zniffer
            for (const freq of freqs.supportedFrequencies) {
                const freqInfo = await this.getFrequencyInfo(freq);
                this._supportedFrequencies.set(freq, freqInfo.frequencyName);
            }
            this.znifferLog.print(`received frequency info:
  current frequency:     ${this._supportedFrequencies.get(freqs.currentFrequency)}
  supported frequencies: ${freqs.supportedFrequencies.map((f) => `\n  · ${f.toString().padStart(2, " ")}: ${this
                ._supportedFrequencies.get(f)}`).join("")}`, "info");
        }
        if (typeof this._options.defaultFrequency === "number"
            && freqs.currentFrequency !== this._options.defaultFrequency
            && this._supportedFrequencies.has(this._options.defaultFrequency)) {
            await this.setFrequency(this._options.defaultFrequency);
        }
        this.emit("ready");
    }
    /**
     * Is called when the serial port has received a Zniffer frame
     */
    serialport_onData(data) {
        let msg;
        try {
            msg = serial_1.ZnifferMessage.from({ data });
        }
        catch (e) {
            console.error(e);
            return;
        }
        if (msg.type === serial_1.ZnifferMessageType.Command) {
            this.handleResponse(msg);
        }
        else {
            const dataMsg = msg;
            const capture = {
                timestamp: new Date(),
                rawData: data,
                frameData: dataMsg.payload,
            };
            this._capturedFrames.push(capture);
            if (this._options.maxCapturedFrames != undefined
                && this._capturedFrames.length > this._options.maxCapturedFrames) {
                this._capturedFrames.shift();
            }
            this.handleDataMessage(dataMsg, capture);
        }
    }
    /**
     * Is called when a Request-type message was received
     */
    handleResponse(msg) {
        // Check if we have a dynamic handler waiting for this message
        for (const entry of this.awaitedMessages) {
            if (entry.predicate(msg)) {
                // We do
                entry.handler(msg);
                return;
            }
        }
    }
    /**
     * Is called when a Request-type message was received
     */
    handleDataMessage(msg, capture) {
        try {
            let convertedRSSI;
            if (this._options.convertRSSI && this._chipType) {
                convertedRSSI = tryConvertRSSI(msg.rssiRaw, this._chipType);
            }
            // Short-circuit if we're dealing with beam frames
            if (msg.frameType === serial_1.ZnifferFrameType.BeamStart
                || msg.frameType === serial_1.ZnifferFrameType.BeamStop) {
                const beam = (0, MPDU_1.parseBeamFrame)(msg);
                beam.frameInfo.rssi = convertedRSSI;
                // Emit the captured frame in a format that's easier to work with for applications.
                this.znifferLog.beam(beam);
                const frame = (0, MPDU_1.beamToFrame)(beam);
                capture.parsedFrame = frame;
                this.emit("frame", frame, capture.frameData);
                return;
            }
            // Only handle messages with a valid checksum, expose the others as CRC errors
            if (!msg.checksumOK) {
                this.znifferLog.crcError(msg);
                const frame = (0, MPDU_1.znifferDataMessageToCorruptedFrame)(msg);
                capture.parsedFrame = frame;
                this.emit("corrupted frame", frame, capture.frameData);
                return;
            }
            const mpdu = (0, MPDU_1.parseMPDU)(msg);
            mpdu.frameInfo.rssi = convertedRSSI;
            // Try to decode the CC while assuming the role of the receiver
            let destSecurityManager;
            let destSecurityManager2;
            let destSecurityManagerLR;
            // Only frames with a destination node id contains something that requires access to the own node ID
            let destNodeId = 0xff;
            let cc;
            // FIXME: Cache data => parsed CC, so we can understand re-transmitted S2 frames
            if (mpdu.payload.length > 0
                && mpdu.headerType !== core_1.MPDUHeaderType.Acknowledgement) {
                if ("destinationNodeId" in mpdu) {
                    destNodeId = mpdu.destinationNodeId;
                    ({
                        securityManager: destSecurityManager,
                        securityManager2: destSecurityManager2,
                        securityManagerLR: destSecurityManagerLR,
                    } = this.getSecurityManagers(mpdu.destinationNodeId));
                }
                // TODO: Support parsing multicast S2 frames
                const ctx = new CCParsingContext_1.ZnifferCCParsingContext(destNodeId, mpdu.homeId, destSecurityManager, destSecurityManager2, destSecurityManagerLR);
                try {
                    cc = cc_1.CommandClass.from(ctx, {
                        data: mpdu.payload,
                        fromEncapsulation: false,
                        nodeId: mpdu.sourceNodeId,
                    });
                }
                catch (e) {
                    // Ignore
                    console.error(e.stack);
                }
            }
            this.znifferLog.mpdu(mpdu, cc);
            // Emit the captured frame in a format that's easier to work with for applications.
            const frame = (0, MPDU_1.mpduToFrame)(mpdu, cc);
            capture.parsedFrame = frame;
            this.emit("frame", frame, capture.frameData);
            // Update the security managers when nonces are exchanged, so we can
            // decrypt the communication
            if (cc?.ccId === core_1.CommandClasses["Security 2"]) {
                const securityManagers = this.getSecurityManagers(mpdu.sourceNodeId);
                const isLR = (0, core_1.isLongRangeNodeId)(mpdu.sourceNodeId)
                    || (0, core_1.isLongRangeNodeId)(destNodeId);
                const senderSecurityManager = isLR
                    ? securityManagers.securityManagerLR
                    : securityManagers.securityManager2;
                const destSecurityManager = isLR
                    ? destSecurityManagerLR
                    : destSecurityManager2;
                if (senderSecurityManager && destSecurityManager) {
                    if (cc instanceof cc_1.Security2CCNonceGet) {
                        // Nonce Get -> all nonces are now invalid
                        senderSecurityManager.deleteNonce(destNodeId);
                        destSecurityManager.deleteNonce(mpdu.sourceNodeId);
                    }
                    else if (cc instanceof cc_1.Security2CCNonceReport && cc.SOS) {
                        // Nonce Report (SOS) -> We only know the receiver's nonce
                        senderSecurityManager.setSPANState(destNodeId, {
                            type: core_1.SPANState.LocalEI,
                            receiverEI: cc.receiverEI,
                        });
                        destSecurityManager.storeRemoteEI(mpdu.sourceNodeId, cc.receiverEI);
                    }
                    else if (cc instanceof cc_1.Security2CCMessageEncapsulation) {
                        const senderEI = cc.getSenderEI();
                        if (senderEI) {
                            // The receiver should now have a valid SPAN state, since decoding the S2 CC updates it.
                            // The security manager for the sender however, does not. Therefore, update it manually,
                            // if the receiver SPAN is indeed valid.
                            const receiverSPANState = destSecurityManager
                                .getSPANState(mpdu.sourceNodeId);
                            if (receiverSPANState.type === core_1.SPANState.SPAN) {
                                senderSecurityManager.setSPANState(destNodeId, receiverSPANState);
                            }
                        }
                    }
                }
            }
            else if (cc?.ccId === core_1.CommandClasses.Security
                && cc instanceof cc_1.SecurityCCNonceReport) {
                const senderSecurityManager = this.getSecurityManagers(mpdu.sourceNodeId).securityManager;
                const destSecurityManager = this.getSecurityManagers(destNodeId).securityManager;
                if (senderSecurityManager && destSecurityManager) {
                    // Both nodes have a shared nonce now
                    senderSecurityManager.setNonce({
                        issuer: mpdu.sourceNodeId,
                        nonceId: senderSecurityManager.getNonceId(cc.nonce),
                    }, {
                        nonce: cc.nonce,
                        receiver: destNodeId,
                    }, { free: true });
                    destSecurityManager.setNonce({
                        issuer: mpdu.sourceNodeId,
                        nonceId: senderSecurityManager.getNonceId(cc.nonce),
                    }, {
                        nonce: cc.nonce,
                        receiver: destNodeId,
                    }, { free: true });
                }
            }
        }
        catch (e) {
            console.error(e);
        }
    }
    /**
     * Waits until a certain serial message is received or a timeout has elapsed. Returns the received message.
     * @param timeout The number of milliseconds to wait. If the timeout elapses, the returned promise will be rejected
     * @param predicate A predicate function to test all incoming messages.
     */
    waitForMessage(predicate, timeout) {
        return new Promise((resolve, reject) => {
            const promise = (0, deferred_promise_1.createDeferredPromise)();
            const entry = {
                predicate,
                handler: (msg) => promise.resolve(msg),
                timeout: undefined,
            };
            this.awaitedMessages.push(entry);
            const removeEntry = () => {
                if (entry.timeout)
                    clearTimeout(entry.timeout);
                const index = this.awaitedMessages.indexOf(entry);
                if (index !== -1)
                    this.awaitedMessages.splice(index, 1);
            };
            // When the timeout elapses, remove the wait entry and reject the returned Promise
            entry.timeout = setTimeout(() => {
                removeEntry();
                reject(new core_1.ZWaveError(`Received no matching message within the provided timeout!`, core_1.ZWaveErrorCodes.Controller_Timeout));
            }, timeout);
            // When the promise is resolved, remove the wait entry and resolve the returned Promise
            void promise.then((cc) => {
                removeEntry();
                resolve(cc);
            });
        });
    }
    async getVersion() {
        const req = new serial_1.ZnifferGetVersionRequest();
        await this.serial?.writeAsync(req.serialize());
        const res = await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferGetVersionResponse, 1000);
        return (0, shared_1.pick)(res, ["chipType", "majorVersion", "minorVersion"]);
    }
    async getFrequencies() {
        const req = new serial_1.ZnifferGetFrequenciesRequest();
        await this.serial?.writeAsync(req.serialize());
        const res = await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferGetFrequenciesResponse, 1000);
        return (0, shared_1.pick)(res, [
            "currentFrequency",
            "supportedFrequencies",
        ]);
    }
    async setFrequency(frequency) {
        const req = new serial_1.ZnifferSetFrequencyRequest({ frequency });
        await this.serial?.writeAsync(req.serialize());
        await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferSetFrequencyResponse, 1000);
        this._currentFrequency = frequency;
    }
    async getFrequencyInfo(frequency) {
        const req = new serial_1.ZnifferGetFrequencyInfoRequest({ frequency });
        await this.serial?.writeAsync(req.serialize());
        const res = await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferGetFrequencyInfoResponse
            && msg.frequency === frequency, 1000);
        return (0, shared_1.pick)(res, ["numChannels", "frequencyName"]);
    }
    /** Starts the capture and discards all previously captured frames */
    async start() {
        if (this.wasDestroyed) {
            throw new core_1.ZWaveError("The Zniffer is not ready or has been destroyed", core_1.ZWaveErrorCodes.Driver_NotReady);
        }
        if (this._active)
            return;
        this._capturedFrames = [];
        this._active = true;
        const req = new serial_1.ZnifferStartRequest();
        await this.serial?.writeAsync(req.serialize());
        await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferStartResponse, 1000);
    }
    async stop() {
        if (!this._active)
            return;
        this._active = false;
        if (!this.serial)
            return;
        const req = new serial_1.ZnifferStopRequest();
        await this.serial?.writeAsync(req.serialize());
        await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferStopResponse, 1000);
    }
    async setBaudrate(baudrate) {
        const req = new serial_1.ZnifferSetBaudRateRequest({ baudrate });
        await this.serial?.writeAsync(req.serialize());
        await this.waitForMessage((msg) => msg instanceof serial_1.ZnifferSetBaudRateResponse, 1000);
    }
    getSecurityManagers(sourceNodeId) {
        if (this.securityManagers.has(sourceNodeId)) {
            return this.securityManagers.get(sourceNodeId);
        }
        // Initialize security
        // Set up the S0 security manager. We can only do that after the controller
        // interview because we need to know the controller node id.
        const S0Key = this._options.securityKeys?.S0_Legacy;
        let securityManager;
        if (S0Key) {
            // this.znifferLog.print(
            // 	"Network key for S0 configured, enabling S0 security manager...",
            // );
            securityManager = new core_1.SecurityManager({
                networkKey: S0Key,
                // FIXME: Track nonces separately for each destination node
                ownNodeId: sourceNodeId,
                nonceTimeout: Number.POSITIVE_INFINITY,
            });
            // } else {
            // 	this.znifferLog.print(
            // 		"No network key for S0 configured, cannot decrypt communication from secure (S0) devices!",
            // 		"warn",
            // 	);
        }
        let securityManager2;
        if (this._options.securityKeys
            // Only set it up if we have security keys for at least one S2 security class
            && Object.keys(this._options.securityKeys).some((key) => key.startsWith("S2_")
                && key in core_1.SecurityClass
                && (0, core_1.securityClassIsS2)(core_1.SecurityClass[key]))) {
            // this.znifferLog.print(
            // 	"At least one network key for S2 configured, enabling S2 security manager...",
            // );
            securityManager2 = new core_1.SecurityManager2();
            // Small hack: Zniffer does not care about S2 duplicates
            securityManager2.isDuplicateSinglecast = () => false;
            // Set up all keys
            for (const secClass of [
                "S2_Unauthenticated",
                "S2_Authenticated",
                "S2_AccessControl",
                "S0_Legacy",
            ]) {
                const key = this._options.securityKeys[secClass];
                if (key) {
                    securityManager2.setKey(core_1.SecurityClass[secClass], key);
                }
            }
            // } else {
            // 	this.znifferLog.print(
            // 		"No network key for S2 configured, cannot decrypt communication from secure (S2) devices!",
            // 		"warn",
            // 	);
        }
        let securityManagerLR;
        if (this._options.securityKeysLongRange?.S2_AccessControl
            || this._options.securityKeysLongRange?.S2_Authenticated) {
            // this.znifferLog.print(
            // 	"At least one network key for Z-Wave Long Range configured, enabling security manager...",
            // );
            securityManagerLR = new core_1.SecurityManager2();
            // Small hack: Zniffer does not care about S2 duplicates
            securityManagerLR.isDuplicateSinglecast = () => false;
            // Set up all keys
            if (this._options.securityKeysLongRange?.S2_AccessControl) {
                securityManagerLR.setKey(core_1.SecurityClass.S2_AccessControl, this._options.securityKeysLongRange.S2_AccessControl);
            }
            if (this._options.securityKeysLongRange?.S2_Authenticated) {
                securityManagerLR.setKey(core_1.SecurityClass.S2_Authenticated, this._options.securityKeysLongRange.S2_Authenticated);
            }
            // } else {
            // 	this.znifferLog.print(
            // 		"No network key for Z-Wave Long Range configured, cannot decrypt Long Range communication!",
            // 		"warn",
            // 	);
        }
        const ret = {
            securityManager,
            securityManager2,
            securityManagerLR,
        };
        this.securityManagers.set(sourceNodeId, ret);
        return ret;
    }
    /** Clears the list of captured frames */
    clearCapturedFrames() {
        this._capturedFrames = [];
    }
    /** Get the captured frames in the official Zniffer application format. */
    getCaptureAsZLFBuffer() {
        // Mimics the current Zniffer software, without using features like sessions and comments
        const header = Buffer.alloc(2048, 0);
        header[0] = 0x68; // zniffer version
        header.writeUInt16BE(0x2312, 0x07fe); // checksum
        return Buffer.concat([
            header,
            ...this._capturedFrames.map(captureToZLFEntry),
        ]);
    }
    /** Saves the captured frames in a `.zlf` file that can be read by the official Zniffer application. */
    async saveCaptureToFile(filePath) {
        await promises_1.default.writeFile(filePath, this.getCaptureAsZLFBuffer());
    }
    /**
     * Terminates the Zniffer instance and closes the underlying serial connection.
     * Must be called under any circumstances.
     */
    async destroy() {
        // Ensure this is only called once and all subsequent calls block
        if (this._destroyPromise)
            return this._destroyPromise;
        this._destroyPromise = (0, deferred_promise_1.createDeferredPromise)();
        this.znifferLog.print("Destroying Zniffer instance...");
        if (this._active) {
            await this.stop().catch(shared_1.noop);
        }
        if (this.serial != undefined) {
            // Avoid spewing errors if the port was in the middle of receiving something
            this.serial.removeAllListeners();
            if (this.serial.isOpen)
                await this.serial.close();
            this.serial = undefined;
        }
        this.znifferLog.print("Zniffer instance destroyed");
        // destroy loggers as the very last thing
        this._logContainer.destroy();
        this._destroyPromise.resolve();
    }
}
exports.Zniffer = Zniffer;
function captureToZLFEntry(capture) {
    const buffer = Buffer.alloc(14 + capture.rawData.length, 0);
    // Convert the date to a .NET datetime
    let ticks = BigInt(capture.timestamp.getTime()) * 10000n
        + 621355968000000000n;
    ticks = ticks | 4000000000000000n; // marks the time as .NET DateTimeKind.Local
    buffer.writeBigUInt64LE(ticks, 0);
    const direction = 0b0000_0000; // inbound, outbound would be 0b1000_0000
    buffer[8] = direction | 0x01; // dir + session ID
    buffer[9] = capture.rawData.length;
    // bytes 10-12 are empty
    capture.rawData.copy(buffer, 13);
    buffer[buffer.length - 1] = 0xfe; // end of frame
    return buffer;
}
//# sourceMappingURL=Zniffer.js.map