"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createMockNodeOptionsFromDump = exports.MockServer = void 0;
const ciao_1 = require("@homebridge/ciao");
const cc_1 = require("@zwave-js/cc");
const core_1 = require("@zwave-js/core");
const mock_1 = require("@zwave-js/serial/mock");
const shared_1 = require("@zwave-js/shared");
const testing_1 = require("@zwave-js/testing");
const deferred_promise_1 = require("alcalzone-shared/deferred-promise");
const node_net_1 = require("node:net");
const Utils_1 = require("./Utils");
class MockServer {
    options;
    constructor(options = {}) {
        this.options = options;
    }
    serialport;
    binding;
    server;
    responder;
    service;
    mockController;
    mockNodes;
    async start() {
        const { port: serialport, binding } = await (0, mock_1.createAndOpenMockedZWaveSerialPort)("/tty/FAKE");
        this.serialport = serialport;
        this.binding = binding;
        console.log("Mock serial port opened");
        // Hook up a fake controller and nodes
        ({ mockController: this.mockController, mockNodes: this.mockNodes } =
            prepareMocks(binding, this.options.config?.controller, this.options.config?.nodes));
        // Call the init hook if it is defined
        if (typeof this.options.config?.onInit === "function") {
            this.options.config.onInit(this.mockController, this.mockNodes);
        }
        // Start a TCP server, listen for connections, and forward them to the serial port
        this.server = (0, node_net_1.createServer)((socket) => {
            if (!this.serialport) {
                console.error("Serial port not initialized");
                socket.destroy();
                return;
            }
            console.log("Client connected");
            socket.pipe(this.serialport);
            this.serialport.on("data", (chunk) => {
                if (typeof chunk === "number") {
                    socket.write(Buffer.from([chunk]));
                }
                else {
                    socket.write(chunk);
                }
            });
            // when the connection is closed, unpipe the streams
            socket.on("close", () => {
                console.log("Client disconnected");
                socket.unpipe(this.serialport);
                this.serialport?.removeAllListeners("data");
            });
        });
        const port = this.options.port ?? 5555;
        this.responder = (0, ciao_1.getResponder)();
        this.service = this.responder.createService({
            name: "zwave-mock-server",
            type: "zwave",
            protocol: "tcp" /* Protocol.TCP */,
            port,
            txt: {
                manufacturer: "Z-Wave JS",
                model: "Mock Server",
            },
        });
        // Do not allow more than one client to connect
        this.server.maxConnections = 1;
        const promise = (0, deferred_promise_1.createDeferredPromise)();
        this.server.on("error", (err) => {
            if (err.code === "EADDRINUSE") {
                promise.reject(err);
            }
        });
        this.server.listen({
            host: this.options.interface,
            port,
        }, async () => {
            const address = this.server.address();
            console.log(`Server listening on tcp://${address.address}:${address.port}`);
            promise.resolve();
            // Advertise the service via mDNS
            try {
                await this.service.advertise();
                console.log(`Enabled mDNS service discovery.`);
            }
            catch (e) {
                console.error(`Failed to enable mDNS service discovery: ${(0, shared_1.getErrorMessage)(e)}`);
            }
        });
    }
    async stop() {
        console.log("Shutting down mock server...");
        await this.service?.end();
        await this.service?.destroy();
        await this.responder?.shutdown();
        this.mockController?.destroy();
        this.server?.close();
        await this.serialport?.close();
        if (this.binding?.isOpen)
            await this.binding?.close();
        console.log("Mock server shut down");
    }
}
exports.MockServer = MockServer;
function prepareMocks(mockPort, controller = {}, nodes = []) {
    const mockController = new testing_1.MockController({
        homeId: 0x7e570001,
        ownNodeId: 1,
        ...controller,
        serial: mockPort,
    });
    // Apply default behaviors that are required for interacting with the driver correctly
    mockController.defineBehavior(...(0, Utils_1.createDefaultMockControllerBehaviors)());
    // Apply custom behaviors
    if (controller.behaviors) {
        mockController.defineBehavior(...controller.behaviors);
    }
    const mockNodes = [];
    for (const node of nodes) {
        const mockNode = new testing_1.MockNode({
            ...node,
            controller: mockController,
        });
        mockController.addNode(mockNode);
        mockNodes.push(mockNode);
        // Apply default behaviors that are required for interacting with the driver correctly
        mockNode.defineBehavior(...(0, Utils_1.createDefaultMockNodeBehaviors)());
        // Apply custom behaviors
        if (node.behaviors) {
            mockNode.defineBehavior(...node.behaviors);
        }
    }
    return {
        mockController,
        mockNodes,
    };
}
function createMockNodeOptionsFromDump(dump) {
    const ret = {
        id: dump.id,
    };
    ret.capabilities = (0, testing_1.getDefaultMockNodeCapabilities)();
    if (typeof dump.isListening === "boolean") {
        ret.capabilities.isListening = dump.isListening;
    }
    if (dump.isFrequentListening !== "unknown") {
        ret.capabilities.isFrequentListening = dump.isFrequentListening;
    }
    if (typeof dump.isRouting === "boolean") {
        ret.capabilities.isRouting = dump.isRouting;
    }
    if (typeof dump.supportsBeaming === "boolean") {
        ret.capabilities.supportsBeaming = dump.supportsBeaming;
    }
    if (typeof dump.supportsSecurity === "boolean") {
        ret.capabilities.supportsSecurity = dump.supportsSecurity;
    }
    if (typeof dump.supportedDataRates === "boolean") {
        ret.capabilities.supportedDataRates = dump.supportedDataRates;
    }
    if (Utils_1.ProtocolVersion[dump.protocol] !== undefined) {
        ret.capabilities.protocolVersion =
            Utils_1.ProtocolVersion[dump.protocol];
    }
    if (dump.deviceClass !== "unknown") {
        ret.capabilities.basicDeviceClass = dump.deviceClass.basic.key;
        ret.capabilities.genericDeviceClass = dump.deviceClass.generic.key;
        ret.capabilities.specificDeviceClass = dump.deviceClass.specific.key;
    }
    ret.capabilities.firmwareVersion = dump.fingerprint.firmwareVersion;
    ret.capabilities.manufacturerId = parseInt(dump.fingerprint.manufacturerId, 16);
    ret.capabilities.productType = parseInt(dump.fingerprint.productType, 16);
    ret.capabilities.productId = parseInt(dump.fingerprint.productId, 16);
    for (const [ccName, ccDump] of Object.entries(dump.commandClasses)) {
        const ccId = core_1.CommandClasses[ccName];
        if (ccId == undefined)
            continue;
        // FIXME: Security encapsulation is not supported yet in mocks
        if (ccId === core_1.CommandClasses.Security
            || ccId === core_1.CommandClasses["Security 2"]) {
            continue;
        }
        ret.capabilities.commandClasses ??= [];
        ret.capabilities.commandClasses.push(createCCCapabilitiesFromDump(ccId, ccDump));
    }
    if (dump.endpoints) {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        for (const [indexStr, endpointDump] of Object.entries(dump.endpoints)) {
            // FIXME: The mocks expect endpoints to be consecutive
            // const index = parseInt(indexStr);
            const epCaps = (0, testing_1.getDefaultMockEndpointCapabilities)(
            // @ts-expect-error We are initializing the device classes above
            ret.capabilities);
            let epCCs;
            if (endpointDump.deviceClass !== "unknown") {
                epCaps.genericDeviceClass =
                    endpointDump.deviceClass.generic.key;
                epCaps.specificDeviceClass =
                    endpointDump.deviceClass.specific.key;
            }
            for (const [ccName, ccDump] of Object.entries(endpointDump.commandClasses)) {
                const ccId = core_1.CommandClasses[ccName];
                if (ccId == undefined)
                    continue;
                // FIXME: Security encapsulation is not supported yet in mocks
                if (ccId === core_1.CommandClasses.Security
                    || ccId === core_1.CommandClasses["Security 2"]) {
                    continue;
                }
                epCCs ??= [];
                epCCs.push(createCCCapabilitiesFromDump(ccId, ccDump));
            }
            ret.capabilities.endpoints ??= [];
            ret.capabilities.endpoints.push({
                ...epCaps,
                commandClasses: epCCs,
            });
        }
    }
    return ret;
}
exports.createMockNodeOptionsFromDump = createMockNodeOptionsFromDump;
function createCCCapabilitiesFromDump(ccId, dump) {
    const ret = {
        ccId,
        isSupported: dump.isSupported,
        isControlled: dump.isControlled,
        secure: dump.secure,
        version: dump.version,
    };
    // Parse CC specific info from values
    if (ccId === core_1.CommandClasses.Configuration) {
        Object.assign(ret, createConfigurationCCCapabilitiesFromDump(dump));
    }
    else if (ccId === core_1.CommandClasses.Notification) {
        Object.assign(ret, createNotificationCCCapabilitiesFromDump(dump));
    }
    return ret;
}
function createConfigurationCCCapabilitiesFromDump(dump) {
    const ret = {
        bulkSupport: false,
        parameters: [],
    };
    for (const val of dump.values) {
        if (typeof val.property !== "number")
            continue;
        // Mocks don't support partial parameters
        if (val.propertyKey != undefined)
            continue;
        // Metadata contains the param information
        if (!val.metadata)
            continue;
        const meta = val.metadata;
        ret.parameters.push({
            "#": val.property,
            valueSize: meta.valueSize ?? 1,
            name: meta.label,
            info: meta.description,
            format: meta.format,
            minValue: meta.min,
            maxValue: meta.max,
            defaultValue: meta.default,
            readonly: !meta.writeable,
        });
    }
    return ret;
}
function createNotificationCCCapabilitiesFromDump(dump) {
    const supportsV1Alarm = findDumpedValue(dump, core_1.CommandClasses.Notification, cc_1.NotificationCCValues.supportsV1Alarm.id, false);
    const ret = {
        supportsV1Alarm,
        notificationTypesAndEvents: {},
    };
    const supportedNotificationTypes = findDumpedValue(dump, core_1.CommandClasses.Notification, cc_1.NotificationCCValues.supportedNotificationTypes.id, []);
    for (const type of supportedNotificationTypes) {
        const supportedEvents = findDumpedValue(dump, core_1.CommandClasses.Notification, cc_1.NotificationCCValues.supportedNotificationEvents(type).id, []);
        ret.notificationTypesAndEvents[type] = supportedEvents;
    }
    return ret;
}
function findDumpedValue(dump, commandClass, valueId, defaultValue) {
    return (dump.values.find((id) => id.property === valueId.property
        && id.propertyKey === valueId.propertyKey)?.value) ?? defaultValue;
}
//# sourceMappingURL=mockServer.js.map