"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.encodeNodeID = exports.parseNodeID = exports.encodeNodeInformationFrame = exports.parseNodeInformationFrame = exports.encodeNodeProtocolInfoAndDeviceClass = exports.parseNodeProtocolInfoAndDeviceClass = exports.encodeNodeProtocolInfo = exports.parseNodeProtocolInfo = exports.NodeType = exports.ProtocolVersion = exports.encodeCCList = exports.parseCCList = exports.encodeCCId = exports.parseCCId = exports.isExtendedCCId = exports.encodeNodeUpdatePayload = exports.parseNodeUpdatePayload = exports.encodeApplicationNodeInformation = exports.parseApplicationNodeInformation = void 0;
const safe_1 = require("@zwave-js/shared/safe");
const consts_1 = require("../consts");
const misc_1 = require("../util/misc");
const CommandClasses_1 = require("./CommandClasses");
function parseApplicationNodeInformation(nif) {
    (0, misc_1.validatePayload)(nif.length >= 2);
    return {
        genericDeviceClass: nif[0],
        specificDeviceClass: nif[1],
        supportedCCs: parseCCList(nif.subarray(2)).supportedCCs,
    };
}
exports.parseApplicationNodeInformation = parseApplicationNodeInformation;
function encodeApplicationNodeInformation(nif) {
    const ccList = encodeCCList(nif.supportedCCs, []);
    return Buffer.concat([
        Buffer.from([nif.genericDeviceClass, nif.specificDeviceClass]),
        ccList,
    ]);
}
exports.encodeApplicationNodeInformation = encodeApplicationNodeInformation;
function parseNodeUpdatePayload(nif, nodeIdType = consts_1.NodeIDType.Short) {
    let offset = 0;
    const { nodeId, bytesRead: nodeIdBytes } = parseNodeID(nif, nodeIdType, offset);
    offset += nodeIdBytes;
    const remainingLength = nif[offset++];
    (0, misc_1.validatePayload)(nif.length >= offset + remainingLength);
    return {
        nodeId,
        basicDeviceClass: nif[offset],
        ...parseApplicationNodeInformation(nif.subarray(offset + 1, offset + remainingLength)),
    };
}
exports.parseNodeUpdatePayload = parseNodeUpdatePayload;
function encodeNodeUpdatePayload(nif, nodeIdType = consts_1.NodeIDType.Short) {
    const ccList = encodeCCList(nif.supportedCCs, []);
    const nodeId = encodeNodeID(nif.nodeId, nodeIdType);
    return Buffer.concat([
        nodeId,
        Buffer.from([
            3 + ccList.length,
            nif.basicDeviceClass,
            nif.genericDeviceClass,
            nif.specificDeviceClass,
        ]),
        ccList,
    ]);
}
exports.encodeNodeUpdatePayload = encodeNodeUpdatePayload;
function isExtendedCCId(ccId) {
    return ccId >= 0xf1;
}
exports.isExtendedCCId = isExtendedCCId;
/**
 * Reads a CC id from the given buffer, returning the parsed CC id and the number of bytes read
 * @param offset The offset at which the CC id is located
 */
function parseCCId(payload, offset = 0) {
    const isExtended = isExtendedCCId(payload[offset]);
    (0, misc_1.validatePayload)(payload.length >= offset + (isExtended ? 2 : 1));
    if (isExtended) {
        return { ccId: payload.readUInt16BE(offset), bytesRead: 2 };
    }
    else {
        return { ccId: payload.readUInt8(offset), bytesRead: 1 };
    }
}
exports.parseCCId = parseCCId;
/**
 * Writes the given CC id into the given buffer at the given location
 * @returns The number of bytes written
 */
function encodeCCId(ccId, payload, offset = 0) {
    if (isExtendedCCId(ccId)) {
        payload.writeUInt16BE(ccId, offset);
        return 2;
    }
    else {
        payload.writeUInt8(ccId, offset);
        return 1;
    }
}
exports.encodeCCId = encodeCCId;
function parseCCList(payload) {
    const ret = {
        supportedCCs: [],
        controlledCCs: [],
    };
    let offset = 0;
    let isAfterMark = false;
    while (offset < payload.length) {
        // Read either the normal or extended ccId
        const { ccId: cc, bytesRead } = parseCCId(payload, offset);
        offset += bytesRead;
        // CCs before the support/control mark are supported
        // CCs after the support/control mark are controlled
        if (cc === CommandClasses_1.CommandClasses["Support/Control Mark"]) {
            isAfterMark = true;
            continue;
        }
        (isAfterMark ? ret.controlledCCs : ret.supportedCCs).push(cc);
    }
    return ret;
}
exports.parseCCList = parseCCList;
function encodeCCList(supportedCCs, controlledCCs) {
    const bufferLength = (0, safe_1.sum)(supportedCCs.map((cc) => (isExtendedCCId(cc) ? 2 : 1)))
        + (controlledCCs.length > 0 ? 1 : 0) // support/control mark
        + (0, safe_1.sum)(controlledCCs.map((cc) => (isExtendedCCId(cc) ? 2 : 1)));
    const ret = Buffer.allocUnsafe(bufferLength);
    let offset = 0;
    for (const cc of supportedCCs) {
        offset += encodeCCId(cc, ret, offset);
    }
    if (controlledCCs.length > 0) {
        ret[offset++] = CommandClasses_1.CommandClasses["Support/Control Mark"];
        for (const cc of controlledCCs) {
            offset += encodeCCId(cc, ret, offset);
        }
    }
    return ret;
}
exports.encodeCCList = encodeCCList;
var ProtocolVersion;
(function (ProtocolVersion) {
    ProtocolVersion[ProtocolVersion["unknown"] = 0] = "unknown";
    ProtocolVersion[ProtocolVersion["2.0"] = 1] = "2.0";
    ProtocolVersion[ProtocolVersion["4.2x / 5.0x"] = 2] = "4.2x / 5.0x";
    ProtocolVersion[ProtocolVersion["4.5x / 6.0x"] = 3] = "4.5x / 6.0x";
})(ProtocolVersion || (exports.ProtocolVersion = ProtocolVersion = {}));
var NodeType;
(function (NodeType) {
    NodeType[NodeType["Controller"] = 0] = "Controller";
    NodeType[NodeType["End Node"] = 1] = "End Node";
})(NodeType || (exports.NodeType = NodeType = {}));
function parseNodeProtocolInfo(buffer, offset, isLongRange = false) {
    (0, misc_1.validatePayload)(buffer.length >= offset + 3);
    const isListening = !!(buffer[offset] & 0b10_000_000);
    const isRouting = !!(buffer[offset] & 0b01_000_000);
    const supportedDataRates = [];
    const speed = buffer[offset] & 0b00_011_000;
    const speedExt = buffer[offset + 2] & 0b111;
    if (isLongRange) {
        // In the LR NIF, the speed bitmask is reserved and contains no information
        // The speedExt bitmask is used instead, but for some reason the bitmask
        // is different from a classic NIF...
        if (speedExt & 0b010) {
            supportedDataRates.push(100000);
        }
    }
    else {
        if (speed & 0b00_010_000) {
            supportedDataRates.push(40000);
        }
        if (speed & 0b00_001_000) {
            supportedDataRates.push(9600);
        }
        if (speedExt & 0b001) {
            supportedDataRates.push(100000);
        }
        if (supportedDataRates.length === 0) {
            supportedDataRates.push(9600);
        }
    }
    const protocolVersion = buffer[offset] & 0b111;
    const capability = buffer[offset + 1];
    const optionalFunctionality = !!(capability & 0b1000_0000);
    let isFrequentListening;
    switch (capability & 0b0110_0000) {
        case 0b0100_0000:
            isFrequentListening = "1000ms";
            break;
        case 0b0010_0000:
            isFrequentListening = "250ms";
            break;
        default:
            isFrequentListening = false;
    }
    const supportsBeaming = !!(capability & 0b0001_0000);
    let nodeType;
    switch (capability & 0b1010) {
        case 0b0010:
            nodeType = NodeType.Controller;
            break;
        case 0b1000:
        // Routing end node
        default:
            // Non-routing end node
            nodeType = NodeType["End Node"];
            break;
    }
    const hasSpecificDeviceClass = !!(capability & 0b100);
    const supportsSecurity = !!(capability & 0b1);
    return {
        isListening,
        isFrequentListening,
        isRouting,
        supportedDataRates,
        protocolVersion,
        optionalFunctionality,
        nodeType,
        supportsSecurity,
        supportsBeaming,
        hasSpecificDeviceClass,
    };
}
exports.parseNodeProtocolInfo = parseNodeProtocolInfo;
function encodeNodeProtocolInfo(info, isLongRange = false) {
    // Technically a lot of these fields are reserved/unused in Z-Wave Long Range,
    // but the only thing where it really matters is the speed bitmask.
    const ret = Buffer.alloc(3, 0);
    // Byte 0 and 2
    if (info.isListening)
        ret[0] |= 0b10_000_000;
    if (info.isRouting)
        ret[0] |= 0b01_000_000;
    if (isLongRange) {
        if (info.supportedDataRates.includes(100000))
            ret[2] |= 0b010;
    }
    else {
        if (info.supportedDataRates.includes(40000))
            ret[0] |= 0b00_010_000;
        if (info.supportedDataRates.includes(9600))
            ret[0] |= 0b00_001_000;
        if (info.supportedDataRates.includes(100000))
            ret[2] |= 0b001;
    }
    ret[0] |= info.protocolVersion & 0b111;
    // Byte 1
    if (info.optionalFunctionality)
        ret[1] |= 0b1000_0000;
    if (info.isFrequentListening === "1000ms")
        ret[1] |= 0b0100_0000;
    else if (info.isFrequentListening === "250ms")
        ret[1] |= 0b0010_0000;
    if (info.supportsBeaming)
        ret[1] |= 0b0001_0000;
    if (info.supportsSecurity)
        ret[1] |= 0b1;
    if (info.nodeType === NodeType["End Node"])
        ret[1] |= 0b1000;
    else
        ret[1] |= 0b0010; // Controller
    if (info.hasSpecificDeviceClass)
        ret[1] |= 0b100;
    return ret;
}
exports.encodeNodeProtocolInfo = encodeNodeProtocolInfo;
function parseNodeProtocolInfoAndDeviceClass(buffer, isLongRange = false) {
    (0, misc_1.validatePayload)(buffer.length >= 5);
    // The specs are a bit confusing here. We parse the response to GetNodeProtocolInfo,
    // which always includes the basic device class, unlike the NIF that was received by
    // the end device. However, the meaning of the flags in the first 3 bytes may change
    // depending on the protocol in use.
    const protocolInfo = parseNodeProtocolInfo(buffer, 0, isLongRange);
    let offset = 3;
    const basic = buffer[offset++];
    const generic = buffer[offset++];
    let specific = 0;
    if (protocolInfo.hasSpecificDeviceClass) {
        (0, misc_1.validatePayload)(buffer.length >= offset + 1);
        specific = buffer[offset++];
    }
    return {
        info: {
            ...protocolInfo,
            basicDeviceClass: basic,
            genericDeviceClass: generic,
            specificDeviceClass: specific,
        },
        bytesRead: offset,
    };
}
exports.parseNodeProtocolInfoAndDeviceClass = parseNodeProtocolInfoAndDeviceClass;
function encodeNodeProtocolInfoAndDeviceClass(info, isLongRange = false) {
    return Buffer.concat([
        encodeNodeProtocolInfo({ ...info, hasSpecificDeviceClass: true }, isLongRange),
        Buffer.from([
            info.basicDeviceClass,
            info.genericDeviceClass,
            info.specificDeviceClass,
        ]),
    ]);
}
exports.encodeNodeProtocolInfoAndDeviceClass = encodeNodeProtocolInfoAndDeviceClass;
function parseNodeInformationFrame(buffer, isLongRange = false) {
    const result = parseNodeProtocolInfoAndDeviceClass(buffer, isLongRange);
    const info = result.info;
    let offset = result.bytesRead;
    let ccList;
    if (isLongRange) {
        const ccListLength = buffer[offset];
        offset += 1;
        (0, misc_1.validatePayload)(buffer.length >= offset + ccListLength);
        ccList = buffer.subarray(offset, offset + ccListLength);
    }
    else {
        ccList = buffer.subarray(offset);
    }
    const supportedCCs = parseCCList(ccList).supportedCCs;
    return {
        ...info,
        supportedCCs,
    };
}
exports.parseNodeInformationFrame = parseNodeInformationFrame;
function encodeNodeInformationFrame(info, isLongRange = false) {
    const protocolInfo = encodeNodeProtocolInfoAndDeviceClass(info, isLongRange);
    let ccList = encodeCCList(info.supportedCCs, []);
    if (isLongRange) {
        ccList = Buffer.concat([Buffer.from([ccList.length]), ccList]);
    }
    return Buffer.concat([protocolInfo, ccList]);
}
exports.encodeNodeInformationFrame = encodeNodeInformationFrame;
function parseNodeID(buffer, type = consts_1.NodeIDType.Short, offset = 0) {
    (0, misc_1.validatePayload)(buffer.length >= offset + type);
    const nodeId = buffer.readUIntBE(offset, type);
    return { nodeId, bytesRead: type };
}
exports.parseNodeID = parseNodeID;
function encodeNodeID(nodeId, type = consts_1.NodeIDType.Short) {
    const ret = Buffer.allocUnsafe(type);
    ret.writeUIntBE(nodeId, 0, type);
    return ret;
}
exports.encodeNodeID = encodeNodeID;
//# sourceMappingURL=NodeInfo.js.map