"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZnifferGetFrequencyInfoResponse = exports.ZnifferGetFrequencyInfoRequest = exports.ZnifferSetBaudRateResponse = exports.ZnifferSetBaudRateRequest = exports.ZnifferStopResponse = exports.ZnifferStopRequest = exports.ZnifferStartResponse = exports.ZnifferStartRequest = exports.ZnifferGetFrequenciesResponse = exports.ZnifferGetFrequenciesRequest = exports.ZnifferSetFrequencyResponse = exports.ZnifferSetFrequencyRequest = exports.ZnifferGetVersionResponse = exports.ZnifferGetVersionRequest = exports.ZnifferDataMessage = exports.ZnifferMessage = void 0;
const core_1 = require("@zwave-js/core");
const shared_1 = require("@zwave-js/shared");
const Constants_1 = require("./Constants");
/**
 * Tests whether the given message constructor options contain a buffer for deserialization
 */
function gotDeserializationOptions(options) {
    return options != undefined && Buffer.isBuffer(options.data);
}
/**
 * Represents a Zniffer message for communication with the serial interface
 */
class ZnifferMessage {
    constructor(
    // public readonly host: ZWaveHost,
    options) {
        // decide which implementation we follow
        if (gotDeserializationOptions(options)) {
            // #1: deserialize from payload
            const payload = options.data;
            // Assume that we're dealing with a complete frame
            this.type = payload[0];
            if (this.type === Constants_1.ZnifferMessageType.Command) {
                this.functionType = payload[1];
                const length = payload[2];
                this.payload = payload.subarray(3, 3 + length);
            }
            else if (this.type === Constants_1.ZnifferMessageType.Data) {
                // The ZnifferParser takes care of segmenting frames, so here we
                // only cut off the type byte from the payload
                this.payload = payload.subarray(1);
            }
            else {
                throw new core_1.ZWaveError(`Invalid Zniffer message type ${this.type}`, core_1.ZWaveErrorCodes.PacketFormat_InvalidPayload);
            }
        }
        else {
            this.type = options.messageType;
            this.functionType = options.functionType;
            this.payload = options.payload || Buffer.allocUnsafe(0);
        }
    }
    type;
    functionType;
    payload; // TODO: Length limit 255
    /** Serializes this message into a Buffer */
    serialize() {
        if (this.type === Constants_1.ZnifferMessageType.Command) {
            return Buffer.concat([
                Buffer.from([
                    this.type,
                    this.functionType,
                    this.payload.length,
                ]),
                this.payload,
            ]);
        }
        else if (this.type === Constants_1.ZnifferMessageType.Data) {
            const ret = Buffer.allocUnsafe(this.payload.length + 1);
            ret[0] = this.type;
            this.payload.copy(ret, 1);
            this.payload[9] = this.payload.length - 10;
            return ret;
        }
        else {
            throw new core_1.ZWaveError(`Invalid Zniffer message type ${this.type}`, core_1.ZWaveErrorCodes.PacketFormat_InvalidPayload);
        }
    }
    /**
     * Retrieves the correct constructor for the next message in the given Buffer.
     * It is assumed that the buffer has been checked beforehand
     */
    static getConstructor(data) {
        const type = data[0];
        // We hardcode the list of constructors here, since the Zniffer protocol has
        // a very limited list of messages
        if (type === Constants_1.ZnifferMessageType.Command) {
            const functionType = data[1];
            switch (functionType) {
                case Constants_1.ZnifferFunctionType.GetVersion:
                    return ZnifferGetVersionResponse;
                case Constants_1.ZnifferFunctionType.SetFrequency:
                    return ZnifferSetFrequencyResponse;
                case Constants_1.ZnifferFunctionType.GetFrequencies:
                    return ZnifferGetFrequenciesResponse;
                case Constants_1.ZnifferFunctionType.Start:
                    return ZnifferStartResponse;
                case Constants_1.ZnifferFunctionType.Stop:
                    return ZnifferStopResponse;
                case Constants_1.ZnifferFunctionType.SetBaudRate:
                    return ZnifferSetBaudRateResponse;
                case Constants_1.ZnifferFunctionType.GetFrequencyInfo:
                    return ZnifferGetFrequencyInfoResponse;
                default:
                    return ZnifferMessage;
            }
        }
        else if (type === Constants_1.ZnifferMessageType.Data) {
            return ZnifferDataMessage;
        }
        else {
            return ZnifferMessage;
        }
    }
    /** Creates an instance of the message that is serialized in the given buffer */
    static from(options) {
        const Constructor = ZnifferMessage.getConstructor(options.data);
        const ret = new Constructor(options);
        return ret;
    }
}
exports.ZnifferMessage = ZnifferMessage;
function computeChecksumXOR(buffer) {
    let ret = 0xff;
    for (let i = 0; i < buffer.length; i++) {
        ret ^= buffer[i];
    }
    return ret;
}
class ZnifferDataMessage extends ZnifferMessage {
    constructor(options) {
        super(options);
        if (gotDeserializationOptions(options)) {
            this.frameType = this.payload[0];
            // bytes 1-2 are 0
            this.channel = this.payload[3] >>> 5;
            this.protocolDataRate = this.payload[3] & 0b11111;
            const checksumLength = this.protocolDataRate >= core_1.ZnifferProtocolDataRate.ZWave_100k
                ? 2
                : 1;
            this.region = this.payload[4];
            this.rssiRaw = this.payload[5];
            if (this.frameType === Constants_1.ZnifferFrameType.Data) {
                core_1.validatePayload.withReason(`ZnifferDataMessage[6] = ${this.payload[6]}`)(this.payload[6] === 0x21);
                core_1.validatePayload.withReason(`ZnifferDataMessage[7] = ${this.payload[7]}`)(this.payload[7] === 0x03);
                // Length is already validated, so we just skip the length byte
                const mpduOffset = 9;
                const checksum = this.payload.readUIntBE(this.payload.length - checksumLength, checksumLength);
                // Compute checksum over the entire MPDU
                const expectedChecksum = checksumLength === 1
                    ? computeChecksumXOR(this.payload.subarray(mpduOffset, -checksumLength))
                    : (0, core_1.CRC16_CCITT)(this.payload.subarray(mpduOffset, -checksumLength));
                this.checksumOK = checksum === expectedChecksum;
                this.payload = this.payload.subarray(mpduOffset, -checksumLength);
            }
            else if (this.frameType === Constants_1.ZnifferFrameType.BeamStart) {
                core_1.validatePayload.withReason(`ZnifferDataMessage[6] = ${this.payload[6]}`)(this.payload[6] === 0x55);
                // There is no checksum
                this.checksumOK = true;
                this.payload = this.payload.subarray(6);
            }
            else if (this.frameType === Constants_1.ZnifferFrameType.BeamStop) {
                // This always seems to contain the same 2 bytes
                // There is no checksum
                this.checksumOK = true;
                this.payload = Buffer.alloc(0);
            }
            else {
                core_1.validatePayload.fail(`Unsupported frame type ${(0, shared_1.getEnumMemberName)(Constants_1.ZnifferFrameType, this.frameType)}`);
            }
        }
        else {
            throw new core_1.ZWaveError(`Sending ${this.constructor.name} is not supported!`, core_1.ZWaveErrorCodes.Driver_NotSupported);
        }
    }
    frameType;
    channel;
    protocolDataRate;
    region;
    rssiRaw;
    checksumOK;
}
exports.ZnifferDataMessage = ZnifferDataMessage;
class ZnifferGetVersionRequest extends ZnifferMessage {
    constructor() {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.GetVersion,
        });
    }
}
exports.ZnifferGetVersionRequest = ZnifferGetVersionRequest;
class ZnifferGetVersionResponse extends ZnifferMessage {
    constructor(options) {
        super(options);
        if (gotDeserializationOptions(options)) {
            this.chipType = (0, core_1.getZWaveChipType)(this.payload[0], this.payload[1]);
            this.majorVersion = this.payload[2];
            this.minorVersion = this.payload[3];
        }
        else {
            throw new core_1.ZWaveError(`Sending ${this.constructor.name} is not supported!`, core_1.ZWaveErrorCodes.Driver_NotSupported);
        }
    }
    chipType;
    majorVersion;
    minorVersion;
}
exports.ZnifferGetVersionResponse = ZnifferGetVersionResponse;
class ZnifferSetFrequencyRequest extends ZnifferMessage {
    constructor(options) {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.SetFrequency,
        });
        this.frequency = options.frequency;
    }
    frequency;
    serialize() {
        this.payload = Buffer.from([this.frequency]);
        return super.serialize();
    }
}
exports.ZnifferSetFrequencyRequest = ZnifferSetFrequencyRequest;
class ZnifferSetFrequencyResponse extends ZnifferMessage {
}
exports.ZnifferSetFrequencyResponse = ZnifferSetFrequencyResponse;
class ZnifferGetFrequenciesRequest extends ZnifferMessage {
    constructor() {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.GetFrequencies,
        });
    }
}
exports.ZnifferGetFrequenciesRequest = ZnifferGetFrequenciesRequest;
class ZnifferGetFrequenciesResponse extends ZnifferMessage {
    constructor(options) {
        super(options);
        if (gotDeserializationOptions(options)) {
            this.currentFrequency = this.payload[0];
            this.supportedFrequencies = [
                ...this.payload.subarray(1),
            ];
        }
        else {
            throw new core_1.ZWaveError(`Sending ${this.constructor.name} is not supported!`, core_1.ZWaveErrorCodes.Driver_NotSupported);
        }
    }
    currentFrequency;
    supportedFrequencies;
}
exports.ZnifferGetFrequenciesResponse = ZnifferGetFrequenciesResponse;
class ZnifferStartRequest extends ZnifferMessage {
    constructor() {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.Start,
        });
    }
}
exports.ZnifferStartRequest = ZnifferStartRequest;
class ZnifferStartResponse extends ZnifferMessage {
}
exports.ZnifferStartResponse = ZnifferStartResponse;
class ZnifferStopRequest extends ZnifferMessage {
    constructor() {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.Stop,
        });
    }
}
exports.ZnifferStopRequest = ZnifferStopRequest;
class ZnifferStopResponse extends ZnifferMessage {
}
exports.ZnifferStopResponse = ZnifferStopResponse;
class ZnifferSetBaudRateRequest extends ZnifferMessage {
    constructor(options) {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.SetBaudRate,
        });
        this.baudrate = options.baudrate;
    }
    baudrate;
    serialize() {
        this.payload = Buffer.from([this.baudrate]);
        return super.serialize();
    }
}
exports.ZnifferSetBaudRateRequest = ZnifferSetBaudRateRequest;
class ZnifferSetBaudRateResponse extends ZnifferMessage {
}
exports.ZnifferSetBaudRateResponse = ZnifferSetBaudRateResponse;
class ZnifferGetFrequencyInfoRequest extends ZnifferMessage {
    constructor(options) {
        super({
            messageType: Constants_1.ZnifferMessageType.Command,
            functionType: Constants_1.ZnifferFunctionType.GetFrequencyInfo,
        });
        this.frequency = options.frequency;
    }
    frequency;
    serialize() {
        this.payload = Buffer.from([this.frequency]);
        return super.serialize();
    }
}
exports.ZnifferGetFrequencyInfoRequest = ZnifferGetFrequencyInfoRequest;
class ZnifferGetFrequencyInfoResponse extends ZnifferMessage {
    constructor(options) {
        super(options);
        if (gotDeserializationOptions(options)) {
            this.frequency = this.payload[0];
            this.numChannels = this.payload[1];
            this.frequencyName = this.payload
                .subarray(2)
                .toString("ascii");
        }
        else {
            throw new core_1.ZWaveError(`Sending ${this.constructor.name} is not supported!`, core_1.ZWaveErrorCodes.Driver_NotSupported);
        }
    }
    frequency;
    numChannels;
    frequencyName;
}
exports.ZnifferGetFrequencyInfoResponse = ZnifferGetFrequencyInfoResponse;
//# sourceMappingURL=ZnifferMessages.js.map