"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.skipBytes = exports.ZnifferParser = void 0;
const node_stream_1 = require("node:stream");
const MessageHeaders_1 = require("../MessageHeaders");
const Constants_1 = require("../message/Constants");
/** Given a buffer that starts with SOF, this method returns the number of bytes the first message occupies in the buffer */
function getMessageLength(data) {
    if (!data || data.length === 0)
        return;
    if (data[0] === MessageHeaders_1.ZnifferMessageHeaders.SOCF) {
        // Control frame: SOF, CMD, remaining length
        if (data.length < 3)
            return;
        const length = data[2];
        return length + 3;
    }
    else if (data[0] === MessageHeaders_1.ZnifferMessageHeaders.SODF) {
        if (data.length < 2)
            return;
        if (data[1] === Constants_1.ZnifferFrameType.BeamStart) {
            // Beam Start frame: SOF, 0x04, 9 other bytes = total 11
            return 11;
        }
        else if (data[1] === Constants_1.ZnifferFrameType.BeamStop) {
            // Beam Stop frame: SOF, 0x05, 5 other bytes = total 7
            return 7;
        }
        // Data frame: SOF, 8 other bytes, remaining length
        if (data.length < 10)
            return;
        const length = data[9];
        return length + 10;
    }
}
class ZnifferParser extends node_stream_1.Transform {
    logger;
    onDiscarded;
    constructor(logger, onDiscarded) {
        // We read byte streams but emit messages
        super({ readableObjectMode: true });
        this.logger = logger;
        this.onDiscarded = onDiscarded;
    }
    receiveBuffer = Buffer.allocUnsafe(0);
    // Allow ignoring the high nibble of an ACK once to work around an issue in the 700 series firmware
    ignoreAckHighNibble = false;
    _transform(chunk, encoding, callback) {
        this.receiveBuffer = Buffer.concat([this.receiveBuffer, chunk]);
        while (this.receiveBuffer.length > 0) {
            // Scan ahead until the next valid byte and log the invalid bytes, if any
            let skip = 0;
            while (skip < this.receiveBuffer.length
                && this.receiveBuffer[skip] !== MessageHeaders_1.ZnifferMessageHeaders.SOCF
                && this.receiveBuffer[skip] !== MessageHeaders_1.ZnifferMessageHeaders.SODF) {
                skip++;
            }
            if (skip > 0) {
                const discarded = this.receiveBuffer.subarray(0, skip);
                this.logger?.discarded(discarded);
                this.onDiscarded?.(discarded);
                // Continue with the next valid byte
                this.receiveBuffer = skipBytes(this.receiveBuffer, skip);
                continue;
            }
            const msgLength = getMessageLength(this.receiveBuffer);
            if (msgLength == undefined || this.receiveBuffer.length < msgLength) {
                // The buffer contains no complete message, we're done here for now
                break;
            }
            else {
                // We have at least one complete message.
                // emit it and slice the read bytes from the buffer
                const msg = this.receiveBuffer.subarray(0, msgLength);
                this.receiveBuffer = skipBytes(this.receiveBuffer, msgLength);
                this.logger?.data("inbound", msg);
                this.push(msg);
            }
        }
        callback();
    }
}
exports.ZnifferParser = ZnifferParser;
/** Skips the first n bytes of a buffer and returns the rest */
function skipBytes(buf, n) {
    return Buffer.from(buf.subarray(n));
}
exports.skipBytes = skipBytes;
//# sourceMappingURL=ZnifferParser.js.map